-- This module provides the datatype of lists.

open import Lib.Proposition
open import Lib.Equality as Eq
open import Lib.Boolean
open import Lib.Nat
open import Lib.Maybe
open import Lib.Inspect

module Lib.List where

-- ----------------------------------------------------------------------
-- * Definition

infixr 7 _::_
infixr 6 _++_

-- The list datatype.
data List (X : Set) : Set where
  nil : List X
  _::_ : X -> List X -> List X

-- A notation for singleton lists.
[[_]] : {X : Set} -> X -> List X
[[ x ]] = x :: nil

-- ----------------------------------------------------------------------
-- * Operations

-- ----------------------------------------------------------------------
-- ** Membership

-- The list membership relation.
data mem {X : Set} (x : X) : List X -> Set where
  mem-head : ∀ {xs} -> mem x (x :: xs)
  mem-tail : ∀ {y xs} -> mem x xs -> mem x (y :: xs)

-- Return the nth member of a list, if any.
nth : {X : Set} -> Nat -> List X -> Maybe X
nth n nil = Nothing
nth zero (x :: xs) = Just x
nth (succ n) (x :: xs) = nth n xs

-- Lemma: The nth member of a list is a member!
lemma-mem-nth : ∀ {X : Set} {x : X} {xs : List X} (n : Nat) -> nth n xs == Just x -> mem x xs
lemma-mem-nth {xs = nil} n ()
lemma-mem-nth {xs = x :: xs} zero refl = mem-head
lemma-mem-nth {xs = x :: xs} (succ n) hyp = mem-tail (lemma-mem-nth n hyp)

-- ----------------------------------------------------------------------
-- ** Map

-- Map a function over all the elements of a list.
map : {X Y : Set} -> (f : X -> Y) -> List X -> List Y
map f nil = nil
map f (x :: xs) = f x :: map f xs

-- Lemma: if x ∈ xs, then f x ∈ map f xs.
lemma-mem-map : {X Y : Set} -> (f : X -> Y) -> (xs : List X) -> ∀ x -> mem x xs -> mem (f x) (map f xs)
lemma-mem-map f nil x ()
lemma-mem-map f (x :: xs) .x mem-head = mem-head
lemma-mem-map f (x :: xs) y (mem-tail hyp) = mem-tail (lemma-mem-map f xs y hyp)

-- ----------------------------------------------------------------------
-- ** Length

-- Return the length of a list.
length : ∀ {X} -> List X -> Nat
length nil = 0
length (x :: xs) = succ (length xs)

-- ----------------------------------------------------------------------
-- ** Appending

-- Append two lists.
_++_ : ∀ {X} -> List X -> List X -> List X
nil ++ ys = ys
(x :: xs) ++ ys = x :: (xs ++ ys)

-- Lemma: Every member of xs is a member of xs ++ ys.
lemma-mem-append1 : ∀ {X} -> (xs ys : List X) -> {x : X} -> mem x xs -> mem x (xs ++ ys)
lemma-mem-append1 nil ys ()
lemma-mem-append1 (x :: xs) ys mem-head = mem-head
lemma-mem-append1 (a :: xs) ys (mem-tail hyp) = mem-tail (lemma-mem-append1 xs ys hyp)

-- Lemma: Every member of ys is a member of xs ++ ys.
lemma-mem-append2 : ∀ {X} -> (xs ys : List X) -> {x : X} -> mem x ys -> mem x (xs ++ ys)
lemma-mem-append2 nil ys hyp = hyp
lemma-mem-append2 (x :: xs) ys hyp = mem-tail (lemma-mem-append2 xs ys hyp)

-- ----------------------------------------------------------------------
-- ** Concatenate a list of lists

-- Concatenate a list of lists.
concat : ∀ {X} -> List (List X) -> List X
concat nil = nil
concat (xs :: xss) = xs ++ concat xss

-- Lemma: any member of a member of xss is a member of (concat xss).
lemma-mem-concat : ∀ {X} -> (xss : List (List X)) -> ∀ {x} {xs} -> mem x xs -> mem xs xss -> mem x (concat xss)
lemma-mem-concat nil mem-x ()
lemma-mem-concat (xs :: xss) mem-x mem-head = lemma-mem-append1 xs (concat xss) mem-x
lemma-mem-concat (y :: xss) mem-x (mem-tail mem-xs) = lemma-mem-append2 y (concat xss) (lemma-mem-concat xss mem-x mem-xs)

-- ----------------------------------------------------------------------
-- ** Reverse a list

-- Append the second list to the reverse of the first one.
reverse-append : ∀ {X} -> List X -> List X -> List X
reverse-append nil ys = ys
reverse-append (x :: xs) ys = reverse-append xs (x :: ys)

-- Reverse the given list.
reverse : ∀ {X} -> List X -> List X
reverse xs = reverse-append xs nil

-- ----------------------------------------------------------------------
-- ** Split

-- Split a list xs into a prefix of length up to n, and a remainder. 
split : ∀ {X} -> Nat -> List X -> List X × List X
split n nil = (nil , nil)
split zero (x :: xs) = (nil , x :: xs)
split (succ n) (x :: xs) with split n xs
... | (xs1 , xs2) = (x :: xs1 , xs2)

-- Lemma: splitting a list and the appending the two parts yields the
-- original list.
lemma-split : ∀ {X} -> (n : Nat) -> (xs : List X) -> ∀ {ys zs} -> split n xs == (ys , zs) -> ys ++ zs == xs
lemma-split n nil refl = refl
lemma-split zero (x :: xs) refl = refl
lemma-split (succ n) (x :: xs) eq with split n xs | lemma-split n xs
lemma-split (succ n) (x :: xs) refl | (xs1 , xs2) | ih = 
   equational (x :: xs1) ++ xs2
           by refl
       equals x :: (xs1 ++ xs2)
           by context (λ □ → x :: □) (ih refl)
       equals x :: xs

-- Like split, but split the list into 3 parts: (x , y , z), where x
-- is of length up to n and y is of length up to m.
split3 : ∀ {X} -> Nat -> Nat -> List X -> List X × List X × List X
split3 n m xs with split n xs
... | (x , ys) with split m ys
... | (y , z) = (x , y , z)

-- Like lemma-split, but for split3.
lemma-split3 : ∀ {X} -> (n m : Nat) -> (xs : List X) -> ∀ {x y z} -> split3 n m xs == (x , y , z) -> x ++ y ++ z == xs
lemma-split3 n m xs hyp with split n xs | inspect (split n) xs
lemma-split3 n m xs hyp | (x , ys) | [ eq1 ] with split m ys | inspect (split m) ys
lemma-split3 n m xs refl | x , ys | [ eq1 ] | y , z | [ eq2 ] =
  equational x ++ y ++ z
          by context (λ □ → x ++ □) hyp2
      equals x ++ ys
          by hyp1
      equals xs
  where
    hyp1 : x ++ ys == xs
    hyp1 = lemma-split n xs eq1

    hyp2 : y ++ z == ys
    hyp2 = lemma-split m ys eq2

-- Span a list according to a predicate p.
span : ∀ {A : Set} -> (p : A -> Boolean) -> List A -> List A × List A
span p nil = nil , nil
span p (x :: xs) with p x
... | false = nil , x :: xs
... | true with span p xs
... | (l , r) = x :: l , r

-- Lemma: put spanned lists together get the original list.
lemma-span : ∀ {A : Set} -> (p : A -> Boolean) -> (xs : List A) -> let sp = span p xs in xs == fst sp ++ snd sp
lemma-span p nil = Eq.refl
lemma-span p (x :: xs) with p x
... | false = Eq.refl
... | true with span p xs | lemma-span p xs
... | x₁ , x₂ | ihp rewrite ihp = Eq.refl


-- ----------------------------------------------------------------------
-- ** Product

-- The product of two lists is the list of all pairs (x , y), where x
-- is taken from X and y is taken from Y.
list-prod : ∀ {X Y : Set} -> List X -> List Y -> List (X × Y)
list-prod xs ys = concat (map (λ y → map (λ x → (x , y)) xs) ys)

-- Lemma: if x and y are members of two lists, then (x , y) is a
-- member of their product.
lemma-mem-list-prod : ∀ {X Y : Set} -> (xs : List X) -> (ys : List Y) -> ∀ {x y} -> mem x xs -> mem y ys -> mem (x , y) (list-prod xs ys)
lemma-mem-list-prod xs ys {x} {y} p q =
  let p' : mem (x , y) (map (λ x → (x , y)) xs)
      p' = lemma-mem-map (λ x -> (x , y)) xs x p

      q' : mem (map (λ x → (x , y)) xs) (map (λ y → map (λ x → (x , y)) xs) ys)
      q' = lemma-mem-map (λ y → map (λ x → x , y) xs) ys y q
  in
    lemma-mem-concat (map (λ y → map (λ x → x , y) xs) ys) p' q'

-- Lemma: xs ++ nil == xs
lemma-append-nil : ∀ {X} (xs : List X) -> xs ++ nil == xs
lemma-append-nil {X} nil = Eq.refl
lemma-append-nil {X} (x :: xs) with lemma-append-nil xs
... | ih rewrite ih = Eq.refl

-- Lemma: concatenation is associative.
lemma-append-assoc : ∀ {X} (xs ys zs : List X) -> (xs ++ ys) ++ zs == xs ++ ys ++ zs
lemma-append-assoc nil ys zs = Eq.refl
lemma-append-assoc (x :: xs) ys zs with lemma-append-assoc xs ys zs
... | ih rewrite ih = Eq.refl

-- Lemma: x :: xs ++ ys == x :: (xs ++ ys).
lemma-::-append : ∀ {X} (x : X) (xs ys : List X) -> x :: xs ++ ys == x :: (xs ++ ys)
lemma-::-append x nil ys = Eq.refl
lemma-::-append x (x₁ :: xs) ys with lemma-::-append x₁ xs ys
... | ih rewrite ih = Eq.refl

-- Lemma: xs ++ y :: ys == (xs ++ y :: nil) ++ ys.
lemma-append-:: : ∀ {X} (xs : List X) (y : X) (ys : List X) -> xs ++ y :: ys == (xs ++ y :: nil) ++ ys
lemma-append-:: nil y ys = Eq.refl
lemma-append-:: (x :: xs) y ys with lemma-append-:: xs y ys
... | ih rewrite ih = Eq.refl


-- ----------------------------------------------------------------------
-- * List quantification

module All where

  infixr 7 _::_

  -- A quantifier for lists. "All P xs" means that all members of xs
  -- have property P.
  data All {A : Set} (P : A -> Set) : List A -> Set where
    [] : All P nil
    _::_ : ∀ {x xs} -> P x -> All P xs -> All P (x :: xs)

  -- A more convenient syntax for list quantification.
  infixr 4 All
  syntax All (λ x -> B) xs = All[ x ∈ xs ] B

  -- The All-elimination rule.
  All-elim : ∀ {X} {x : X} {xs P} -> mem x xs -> All P xs -> P x
  All-elim mem-head (x :: a) = x
  All-elim (mem-tail m) (y :: a) = All-elim m a

-- ----------------------------------------------------------------------
-- * Strict functions

module Strict where

  -- Apply a function to a list, after first expanding the list
  -- strictly. This is basically a trick to work around certain
  -- problems with Agda's evaluation mechanism. Agda's default
  -- mechanism for evaluating expressions can be very inefficient, and
  -- using 'strict' can occasionally speed things up.
  strict : {X : Set} {P : List X -> Set} -> (xs : List X) -> (k : (xs : List X) -> P xs) -> P xs
  strict nil k = k nil
  strict (x :: xs) k = strict xs (\xs' -> k (x :: xs'))

  lemma-strict : {X : Set} {P : List X -> Set} -> (xs : List X) -> (k : (xs : List X) -> P xs) -> strict xs k == k xs
  lemma-strict nil k = refl
  lemma-strict (x :: xs) k = lemma-strict xs (\xs' -> k (x :: xs'))

-- ----------------------------------------------------------------------
-- * Additional properties of lists

-- If two non-empty lists are equal, then so are their heads.
lemma-list-cancellation-head : ∀ {X} {x y : X} {xs ys : List X} -> x List.:: xs == y :: ys -> x == y
lemma-list-cancellation-head refl = refl

-- If two non-empty lists are equal, then so are their tails.
lemma-list-cancellation-tail : ∀ {X} {x y : X} {xs ys : List X} -> x List.:: xs == y :: ys -> xs == ys
lemma-list-cancellation-tail refl = refl

-- ----------------------------------------------------------------------
-- * Some support for proofs by case distinction

is-L-list : ∀ {A B} -> List (A ∪ B) -> Set
is-L-list {A} {B} xs = (∃ λ (xs' : List A) -> map ⌜_ xs' == xs)

is-R-list : ∀ {A B} -> List (A ∪ B) -> Set
is-R-list {A} {B} xs = (∃ λ (xs' : List B) -> map _⌝ xs' == xs)

take-while-isL : ∀ {A B} -> (xs : List (A ∪ B)) -> List (A ∪ B)
take-while-isL nil = nil
take-while-isL (⌜ x :: xs) with take-while-isL xs
... | xs' = ⌜ x :: xs'
take-while-isL (x ⌝ :: xs) = nil

take-while-isR : ∀ {A B} -> (xs : List (A ∪ B)) -> List (A ∪ B)
take-while-isR nil = nil
take-while-isR (x ⌝ :: xs) with take-while-isR xs
... | xs' = x ⌝ :: xs'
take-while-isR (⌜ x :: xs) = nil

drop-while-isL : ∀ {A B} -> (xs : List (A ∪ B)) -> List (A ∪ B)
drop-while-isL nil = nil
drop-while-isL (⌜ x :: xs) = drop-while-isL xs
drop-while-isL (x ⌝ :: xs) = x ⌝ :: xs

drop-while-isR : ∀ {A B} -> (xs : List (A ∪ B)) -> List (A ∪ B)
drop-while-isR nil = nil
drop-while-isR (x ⌝ :: xs) = drop-while-isR xs
drop-while-isR (⌜ x :: xs) = ⌜ x :: xs

lemma-take-drop-isL : ∀ {A B} -> (xs : List (A ∪ B)) -> take-while-isL xs ++ drop-while-isL xs == xs
lemma-take-drop-isL nil = refl
lemma-take-drop-isL (⌜ x :: xs) with lemma-take-drop-isL xs
... | ih = cong _::_ refl ih
lemma-take-drop-isL (x ⌝ :: xs) with lemma-take-drop-isL xs
... | ih = refl

lemma-take-drop-isR : ∀ {A B} -> (xs : List (A ∪ B)) -> take-while-isR xs ++ drop-while-isR xs == xs
lemma-take-drop-isR nil = refl
lemma-take-drop-isR (x ⌝ :: xs) with lemma-take-drop-isR xs
... | ih = cong _::_ refl ih
lemma-take-drop-isR (⌜ x :: xs) with lemma-take-drop-isR xs
... | ih = refl

lemma-take-while-isL-isL : ∀ {A B} -> (xs : List (A ∪ B)) -> is-L-list (take-while-isL xs)
lemma-take-while-isL-isL nil = (nil , refl)
lemma-take-while-isL-isL (⌜ x :: xs) with lemma-take-while-isL-isL xs
... | x₁ , x₂ = x :: x₁ , cong _::_ refl x₂
lemma-take-while-isL-isL (x ⌝ :: xs) = nil , refl


lemma-take-while-isR-isR : ∀ {A B} -> (xs : List (A ∪ B)) -> is-R-list (take-while-isR xs)
lemma-take-while-isR-isR nil = (nil , refl)
lemma-take-while-isR-isR (x ⌝ :: xs) with lemma-take-while-isR-isR xs
... | x₁ , x₂ = x :: x₁ , cong _::_ refl x₂
lemma-take-while-isR-isR (⌜ x :: xs) = nil , refl


lift-L' : ∀ {A B} -> Nat -> (f : List A -> List A) -> List (A ∪ B) -> List (A ∪ B)
lift-L' zero f xs = xs
lift-L' (succ n) f nil = nil
lift-L' (succ n) f a@(⌜ x :: xs) = map ⌜_ (f (fst (lemma-take-while-isL-isL a))) ++ lift-L' n f (drop-while-isL a) 
lift-L' (succ n) f (x ⌝ :: xs) = x ⌝ :: lift-L' (succ n) f xs

lift-L : ∀ {A B} -> (f : List A -> List A) -> List (A ∪ B) -> List (A ∪ B)
lift-L f xs = lift-L' (length xs) f xs


lift-R' : ∀ {A B} -> Nat -> (f : List B -> List B) -> List (A ∪ B) -> List (A ∪ B)
lift-R' zero f xs = xs
lift-R' (succ n) f nil = nil
lift-R' (succ n) f a@(x ⌝ :: xs) = map _⌝ (f (fst (lemma-take-while-isR-isR a))) ++ lift-R' n f (drop-while-isR a) 
lift-R' (succ n) f (⌜ x :: xs) = ⌜ x :: lift-R' (succ n) f xs

lift-R : ∀ {A B} -> (f : List B -> List B) -> List (A ∪ B) -> List (A ∪ B)
lift-R f xs = lift-R' (length xs) f xs

-- ----------------------------------------------------------------------
-- ** Finiteness

module Finite where

  -- A set X is finite iff there is a list having all elements of X as
  -- members.
  Finite : (X : Set) -> Set
  Finite X = ∃ λ (xs : List X) -> ∀ x -> mem x xs
  
  -- Lemma: The product of two finite sets is finite.
  lemma-finite-product : {X Y : Set} -> Finite X -> Finite Y -> Finite (X × Y)
  lemma-finite-product (xs , x-enum) (ys , y-enum) = (zs , z-enum)
    where
      zs = concat (map (λ y → map (λ x → (x , y)) xs) ys)
  
      z-enum : ∀ z -> mem z zs
      z-enum (x , y) = lemma-mem-list-prod xs ys (x-enum x) (y-enum y)

-- ----------------------------------------------------------------------
-- ** Case distinction over a finite set

module By-Cases where

  open Finite
  open All

  -- When X is finite, we can prove ∀-statements by finite case
  -- distinction.
  by-finite-cases : {X : Set} -> (fin : Finite X) -> (x : X) -> (B : X -> Set) -> All[ x ∈ fst fin ] B x -> B x
  by-finite-cases (xs , enum) x B hyp = All-elim (enum x) hyp
  
  -- Equality of lists implies a list of equalities.
  lemma-list-equality : {X Y : Set} -> (xs : List X) -> (eq : X -> Y × Y) -> let eqs = map eq xs in map fst eqs == map snd eqs -> All[ x ∈ xs ] fst (eq x) == snd (eq x)
  lemma-list-equality nil eq hyp = []
  lemma-list-equality (x :: xs) eq hyp = lemma-list-cancellation-head hyp :: (lemma-list-equality xs eq (lemma-list-cancellation-tail hyp))
  
  -- Reduce a finite family of equalities to a single equality.
  equality-by-cases : {X Y : Set} -> (fin : Finite X) -> (x : X) -> (eq : X -> Y × Y) ->
                      let xs = fst fin
                          eqs = map eq xs
                      in
                      map fst eqs == map snd eqs -> fst (eq x) == snd (eq x)
  equality-by-cases fin x eq hyp =
    let xs = fst fin
        enum = snd fin
        eqs = map eq xs
    in
      by-finite-cases fin x (λ x → fst (eq x) == snd (eq x)) (lemma-list-equality xs eq hyp)


